#!/usr/bin/env python

"""label_ui.py
======== Starting the UI ======== 

Run this script inside the root directory of the repository.
python label_ui.py <model_name> <resolution> [corner_file_version]

<> entries are mandatory.  

[] entry is optional. If it is not supplied, then no annotations will
be shown. If you want to view the annotations used in the paper,
supply corner_file_version=1.

Examples:

python label_ui.py axe 32 
Open the polygon in ``axe/'' with resolution 32.  When the code
executes no corners are selected at the beginning

python label_ui.py bow 64 1
Open the polygon in ``bow/'' with resolution 64.  When the code
executes the corners from the file bow/64_corners.1.txt are selected.

======== Controls  ======== 

Click on a point to select/deselect it.
Press ``o'' to use the zoom tool.
Press ``h'' to go back to revert the zoom.
Press ``a'' to save the labelling
Press ``r'' to clear all annotations
Press ``v'' to toggle viewing the point numbers
Press ``up, pageup'' to increase point size
Press ``down, pagedown'' to reduce point size

======= Where are the files stored =======

All the geometry files are stored in:
<model_name>/<resolution>_boundary.txt
Where the format is:
<point_x> <point_y> 

A .png picture is also provided in
<model_name>/<resolution>_image.png

The UI tries to save the corners in:
<model_name>/<resolution>_corners.1.txt

But if that file exists, then it tries to save in:
<model_name>/<resolution>_corners.2.txt
and so on.

"""

from __future__ import print_function

import os
import sys
import math
import numpy as np
import matplotlib.pyplot as plt


## --------------------------------------------------------------------------
##  Converts x.y to x.1.y if x.1.y does not exist
##  Otherwise converts it to x.2.y
## --------------------------------------------------------------------------
def append_number_to_name(filename):

    candid = filename.rsplit('.',2)[-2]

    ## candid is not a number
    if( not candid.isdigit() ):
        #print 'bad %s %s' % (candid, candid[-1])
        all_str = filename.rsplit('.',1)
        filename_out = "%s.1.%s" % (all_str[0], all_str[1])
        return append_number_to_name(filename_out)
    
   ## the file already exists     
    if os.path.isfile(filename):
        # print 'bad %s exists' % candid
        all_str = filename.rsplit('.',2)
        filename_out = "%s.%d.%s" % (all_str[0], int(all_str[1])+1, all_str[2])
        return append_number_to_name(filename_out)

    ## everything is fine
    return filename

## --------------------------------------------------------------------------
##  write_array_to_file
##  Write an array to a file.
## --------------------------------------------------------------------------
def write_array_to_file(filename, data):
    filename = append_number_to_name(filename)
    f = open(filename, 'w')  
    for d in data:
        f.write("%d\n" % d)

    f.close()
    print("Wrote data to %s" % filename)


## --------------------------------------------------------------------------
##  read_lines_from_file
##  Read all the lines of a file.
## --------------------------------------------------------------------------
def read_lines_from_file(fname):
    try:
        f = open(fname, 'r') 
        lines = f.readlines()
        f.close()
        return lines;
    except:
        assert False, "Could not open file ``{}'' ".format(fname)


## --------------------------------------------------------------------------
##  read_samples_from_file
##  Reads a boundary file.
##  Returns:
##    pts --- a n_points * 4 matrix list[list]
##    where columns are x, y, x+normal_x, y+normal_x
##    types --- a string array containing point types
## --------------------------------------------------------------------------
def read_samples_from_file(filename):
    lines = read_lines_from_file(filename)
    pts = []
    for line in lines:
        xy = line.strip().split() # maximum of 2 splits
        xy = [val for val in xy if val != '' ]
        pts.append([float(xy[0]), float(xy[1])])
    pts = np.array(pts)
    return pts

## --------------------------------------------------------------------------
##  Global variables
## --------------------------------------------------------------------------

class GlobalStruct:
    _old_paper_color = (float(248)/255,float(236)/255,float(194)/255)
    _nice_website_color = (float(224)/255,float(199)/255,float(178)/255)

    def __init__(self):
        ## The id of the selected corners
        self.output_id = []

        ## The name of the output file
        self.output_name = "xxxxx"

        ## Color of the corners
        self.colors = []

        ## The scatter object
        self.scatter = []

        ## The figure object
        self.fig  = None
        self.ax = None

        ## Corner annotations
        self.annotations = None

    def init_fig(self):
        self.fig, self.ax = plt.subplots(nrows=1, ncols=1)
        self.ax.set_facecolor(self._nice_website_color)    
        self.fig.set_facecolor(self._nice_website_color)    

    def reinit_colors(self):
        # print self.colors
        self.scatter.set_facecolors(self.colors)
        self.fig.canvas.draw()

    def is_corner_present(self, idx):
        return idx in self.output_id
    
    def add_corner(self, idx):
        self.output_id.append( idx )
        self.colors[idx] = 'g'
        print(self.output_id)
        self.reinit_colors()

    def remove_corner(self, idx):
        self.output_id.remove( idx )
        self.colors[idx] = 'r'
        print(self.output_id)
        self.reinit_colors()

    def remove_corners(self):
        self.output_id = []
        for ii in range(0, len(self.colors)):
            self.colors[ii] = 'r'
        self.reinit_colors()

    def toggle_annotation_visibility(self):
        for ann in self.annotations:
            ann.set_visible(not ann.get_visible() )
        self.fig.canvas.draw()

    def increase_scatter_size(self, val):
        sizes = self.scatter.get_sizes()
        # print( sizes )
        sizes = sizes + val
        # print( sizes )
        sizes = self.scatter.set_sizes(sizes)
        self.fig.canvas.draw()
    
    def decrease_scatter_size(self, val):
        sizes = self.scatter.get_sizes()
        if(sizes[0] - val < 0): return
        # print( sizes )
        sizes = sizes - val
        # print( sizes )
        sizes = self.scatter.set_sizes(sizes)
        self.fig.canvas.draw()
                    
global_struct = GlobalStruct()

## --------------------------------------------------------------------------
##  Events for matplotlib
## --------------------------------------------------------------------------
def event_press(event):
    global global_struct
    ## print('press', event.key)
    
    if event.key == 'a':
        global_struct.output_id.sort()
        write_array_to_file(global_struct.output_name, global_struct.output_id)
    elif event.key == 'r':
        global_struct.remove_corners()
    elif event.key == 'v':
        global_struct.toggle_annotation_visibility()
    elif event.key == 'pageup':
        global_struct.increase_scatter_size(5)
    elif event.key == 'pagedown':
        global_struct.decrease_scatter_size(5)
    elif event.key == 'up':
        global_struct.increase_scatter_size(1)
    elif event.key == 'down':
        global_struct.decrease_scatter_size(1)
    # else:
    #     print(event.key)

def event_onpick(event):
    global global_struct
    
    ## print('onpick scatter:', event.ind)
    index = event.ind[0]
    if not global_struct.is_corner_present( index):
        global_struct.add_corner( index )
    else:
        global_struct.remove_corner( index )
        
## --------------------------------------------------------------------------
##           Main function
## --------------------------------------------------------------------------
def main():
    global global_struct

    ## Read the input
    corner_preload_id = None
    resolution = None
    model_name = None
    try:
        if len(sys.argv) == 3:
            model_name = sys.argv[1]
            resolution = int(sys.argv[2])
            corner_preload_id = ''
        elif len(sys.argv) == 4:
            model_name = sys.argv[1]
            resolution = int(sys.argv[2])
            corner_preload_id= sys.argv[3]
        else:
            print("Invalid number of inputs %d \n" % len(sys.argv))
            print(__doc__)
            return
    except:
        print("Invalid input \n")
        print(__doc__)
        return
    

    ## Create the figure and axis
    global_struct.init_fig()
    axes = global_struct.ax 

    ## Read the point locations
    model_path = "./%s/%d_boundary.txt" % (model_name,resolution)
    global_struct.output_name =  ("./%s/%d_corners.txt" %
                                  (model_name,resolution) )
    samples = read_samples_from_file(model_path)
    n_samples = samples.shape[ 0 ]
    global_struct.colors = ['r'] * n_samples

    ## draw the edges
    for ii in range(n_samples):
        
        jj = (ii+1) % n_samples
        xx = [samples[ii,0], samples[jj,0]]
        yy = [samples[ii,1], samples[jj,1]]
        
        axes.plot(xx, yy,
                  color='b', linestyle='-', linewidth=3,
                  alpha=1, zorder=0.5)

    ## draw the points
    global_struct.scatter =  axes.scatter(samples[:, 0],
                                          samples[:, 1],
                                          c=global_struct.colors,
                                          picker=True, zorder=3)

    ## Annotate corners
    annotations = []
    for i in range(0, samples.shape[0]):
        ann = axes.annotate(str(i), samples[i,:2]+[0.1,0.1],
                            size=9, zorder=16, color='k' )
        annotations.append(ann)
    global_struct.annotations = annotations
    global_struct.toggle_annotation_visibility()


    ## If 0 given as corner preload, run the file without extension
    ## A not so nice hack.
    corner_preload_file_name = None
    if (corner_preload_id == '0'):
        corner_preload_file_name = ('./%s/%d_corners.txt' %
                                    (model_name, resolution))
    ## Otherwise, load the one with the given extension        
    elif not (corner_preload_id == ''):
        corner_preload_file_name = ('./%s/%d_corners.%s.txt' %
                                    (model_name, resolution, corner_preload_id))
    if(corner_preload_file_name is not None):
        try:
            lines = read_lines_from_file(corner_preload_file_name)
            for l in lines:
                global_struct.add_corner( int(l) )
        except:
            assert 0, "Bad formatting in {}".format(corner_preload_file_name)


    ## set the events
    global_struct.fig.canvas.mpl_connect('pick_event', event_onpick)
    global_struct.fig.canvas.mpl_connect('key_press_event', event_press)
    global_struct.fig.canvas.set_window_title('model: %s, resolution: %d' % (model_name, resolution))

    ## set graph properties
    limits = resolution*(1.5);
    axes.set_xlim(0, limits)
    axes.set_ylim(limits, 0)
    axes.set_aspect('equal')
    axes.axis('equal')
    plt.show()

## --------------------------------------------------------------------------
##           Main sentinel
## --------------------------------------------------------------------------
if __name__ == "__main__":
    main()
    


